/*
    generate_simple.c - Part of psiconv, a PSION 5 file formats converter
    Copyright (c) 2000-2004  Frodo Looijaard <frodol@dds.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "config.h"
#include "compat.h"

#include <string.h>
#include <stdlib.h>

#include "generate_routines.h"
#include "error.h"

#ifdef DMALLOC
#include <dmalloc.h>
#endif

static int psiconv_write_string_aux(const psiconv_config config,
                                    psiconv_buffer buf, int lev,
				    const psiconv_string_t value,int kind);

int psiconv_write_u8(const psiconv_config config,psiconv_buffer buf,
                     int lev,const psiconv_u8 value)
{
  psiconv_progress(config,lev,0,"Writing u8");
  return psiconv_buffer_add(buf,value);
}

int psiconv_write_u16(const psiconv_config config,psiconv_buffer buf,
                      int lev,const psiconv_u16 value)
{
  int res;
  psiconv_progress(config,lev,0,"Writing u16");
  if ((res = psiconv_buffer_add(buf,value & 0xff))) 
	return res;
  return psiconv_buffer_add(buf,(value & 0xff00) >> 8);
}

int psiconv_write_u32(const psiconv_config config,psiconv_buffer buf,
                      int lev,const psiconv_u32 value)
{
  int res;
  psiconv_progress(config,lev,0,"Writing u32");
  if ((res = psiconv_buffer_add(buf,value & 0xff))) 
	return res;
  if ((res = psiconv_buffer_add(buf,(value & 0xff00) >> 8))) 
	return res;
  if ((res = psiconv_buffer_add(buf,(value & 0xff0000) >> 16))) 
	return res;
  return psiconv_buffer_add(buf,(value & 0xff000000) >> 24);
}

int psiconv_write_S(const psiconv_config config,psiconv_buffer buf, 
                    int lev,const psiconv_u32 value)
{
  psiconv_progress(config,lev,0,"Writing S");
  if (value < 0x40) 
    return psiconv_write_u8(config,buf,lev+1,value * 4 + 2);
  else if (value < 0x2000) 
    return psiconv_write_u16(config,buf,lev+1,value * 8 + 3);
  else {
    psiconv_error(config,0,psiconv_buffer_length(buf),
                 "Don't know how to write S value larger than 0x2000 "
                 "(trying %x)",value);
    return -PSICONV_E_GENERATE;
  }
}

int psiconv_write_X(const psiconv_config config,psiconv_buffer buf,
                    int lev, const psiconv_u32 value)
{
  psiconv_progress(config,lev,0,"Writing X");
  if (value < 0x80) 
    return psiconv_write_u8(config,buf,lev+1,value * 2);
  else if (value < 0x4000) 
    return psiconv_write_u16(config,buf,lev+1,value * 4 + 1);
  else if (value < 0x20000000) 
    return psiconv_write_u16(config,buf,lev+1,value * 8 + 3);
  else {
    psiconv_error(config,0,psiconv_buffer_length(buf),
                 "Don't know how to write X value larger than 0x20000000 "
                 "(trying %x)",value);
    return -PSICONV_E_GENERATE;
  }
}

int psiconv_write_length(const psiconv_config config,psiconv_buffer buf, 
                         int lev,const psiconv_length_t value)
{
  psiconv_progress(config,lev,0,"Writing length");
  return psiconv_write_u32(config,buf,lev+1,value * (1440.0/2.54) + 0.5);
}

int psiconv_write_size(const psiconv_config config,psiconv_buffer buf,
                       int lev, psiconv_size_t value)
{
  psiconv_progress(config,lev,0,"Writing size");
  return psiconv_write_u32(config,buf,lev+1,value * 20.0 + 0.5);
}

int psiconv_write_bool(const psiconv_config config,psiconv_buffer buf, 
                       int lev,const psiconv_bool_t value)
{
  psiconv_progress(config,lev,0,"Writing bool");
  if ((value != psiconv_bool_true) && (value != psiconv_bool_false)) 
    psiconv_warn(config,0,psiconv_buffer_length(buf),
                 "Boolean has non-enum value (found %d)",value);
  return psiconv_write_u8(config,buf,lev+1,value == psiconv_bool_false?0:1);
}

int psiconv_write_string(const psiconv_config config,psiconv_buffer buf,
                         int lev, const psiconv_string_t value)
{
  psiconv_progress(config,lev,0,"Writing string");
  return psiconv_write_string_aux(config,buf,lev+1,value,-1);
}

int psiconv_write_short_string(const psiconv_config config,psiconv_buffer buf, 
                               int lev,const psiconv_string_t value)
{
  psiconv_progress(config,lev,0,"Writing short string");
  return psiconv_write_string_aux(config,buf,lev+1,value,-2);
}

int psiconv_write_charlist(const psiconv_config config,psiconv_buffer buf, 
                           int lev,const psiconv_string_t value)
{
  psiconv_progress(config,lev,0,"Writing short string");
  return psiconv_write_string_aux(config,buf,lev+1,value,0);
}


int psiconv_write_string_aux(const psiconv_config config,psiconv_buffer buf,
                             int lev, const psiconv_string_t value,int kind)
{
  int res,i,len;

  psiconv_progress(config,lev,0,"Writing string (auxiliary)");
  len = psiconv_unicode_strlen(value);
  if (!value) {
    psiconv_error(config,0,psiconv_buffer_length(buf),
                 "NULL string");
    return -PSICONV_E_GENERATE;
  }
  if (kind == -1) 
    res = psiconv_write_S(config,buf,lev+1,len);
  else if (kind == -2)
    res = psiconv_write_u8(config,buf,lev+1,len);
  else
    res = 0;
  if (res)
    return res;

  for (i = 0; i < len; i++) 
    if ((res = psiconv_unicode_write_char(config,buf,lev+1,value[i]))) 
      return res;
  return -PSICONV_E_OK;
}

int psiconv_write_offset(const psiconv_config config,psiconv_buffer buf, 
                         int lev,psiconv_u32 id)
{
  psiconv_progress(config,lev,0,"Writing offset");
  return psiconv_buffer_add_reference(buf,id);
}
